# coding=utf-8
from __future__ import division
from builtins import map
from builtins import range
from builtins import object
import logging, sys, bisect
from datetime import datetime
import re, traceback
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit
import ase.geometry
import ase.data
import ase.build
import json
from ase import Atoms
from ase.data import atomic_numbers, chemical_symbols
from math import pi
from contextlib import contextmanager

@contextmanager
def open_section(parser, backend, name):
    parser.gidSections[name] = backend.openSection(name)
    yield parser.gidSections[name]
    backend.closeSection(name, parser.gidSections[name])

def metaNameConverter(keyName):
    newName = keyName.lower().replace(" ", "").replace("-", "")
    newName = newName.replace("(", "").replace(")", "")
    newName = newName.replace("[", "").replace("]", "")
    newName = newName.replace(",", "").replace(".", "")
    newName = newName.replace("\\", "").replace("/", "")
    newName = newName.replace("'", "").replace(":", "")
    return newName

def metaNameConverter_UnderscoreAll(keyName):
    newName = keyName.lower()
    newName = ' '.join(newName.split())
    newName = newName.replace(" ", "_").replace("-", "_").replace(".", "_")
    newName = metaNameConverter(newName)
    return newName

def metaNameConverter_OpenKIM(keyName):
    newName = metaNameConverter_UnderscoreAll(keyName)
    newName = 'x_openkim_' + newName
    return newName

def secondsFromEpoch(date):
    epoch = datetime(1970,1,1)
    ts=date-epoch
    return ts.seconds + ts.microseconds/1000.0

def KIMQueryReader(jsonfile):
    with open(jsonfile, 'r') as dbfile:
        data = json.load(dbfile)
    return data['QUERY']

class OpenkimContext(object):
    def __init__(self):
        self.parser = None
        self.weights = None
        self.lastSystemDescription = None
        self.labels = None
        self.singleConfCalcs = []
        self.KIM_TE = 0
        self.KIM_TD = None
        self.KIM_MO = None
        self.KIM_MD = None
        self.cell = None
        self.temperature = []
        self.cohesivepot = []
        self.cohesiveeng = []

    def reset(self):
        self.lastSystemDescription = None
        self.labels = None
        self.singleConfCalcs = []
        self.KIM_TE = 0
        self.KIM_TD = None
        self.KIM_MO = None
        self.KIM_MD = None
        self.cell = None
        self.temperature = []
        self.cohesivepot = []
        self.cohesiveeng = []

    def startedParsing(self, parser):
        self.parser = parser

    def onEnd_program(self, parser, querydict):
        backend = parser.backend
        backend.addValue("program_name", "OpenKIM")
        backend.addValue("program_version", g(querydict, "meta.runner.short-id", ""))
        date = g(querydict, "meta.created_on")
        pdate = None
        if date:
            pdate = datetime.strptime(date.strip(), "%Y-%m-%d %H:%M:%S.%f")
        if pdate:
            backend.addValue("program_compilation_datetime", secondsFromEpoch(pdate))

    def onEnd_openkim_data(self, parser, kimquery):
        for k,v in kimquery.items():
            nomad_k = metaNameConverter_OpenKIM(k)
            if isinstance(v, (list,tuple)):
                dictInList = False
                if len(v)>0:
                    if isinstance(v[0], dict):
                        dictInList = True
                if dictInList:
                    for it, item in enumerate(v):
                        if isinstance(item, dict):
                            #keyFooter = '_' + str(it+1)
                            for key, val in item.items():
                                #key = nomad_k + '_' + key + keyFooter
                                key = nomad_k + '_' + key
                                if isinstance(val, list):
                                    backend.addArrayValues(key,np.asarray(val))
                                else:
                                    backend.addValue(key,val)
                else:
                    backend.addArrayValues(nomad_k,np.asarray(v))
            elif isinstance(v, dict):
                for key, val in item.items():
                    key = nomad_k + '_' + key
                    if isinstance(val, list):
                        backend.addArrayValues(key,np.asarray(val))
                    else:
                        backend.addValue(key,val)
            else:
                backend.addValue(nomad_k,v)

    def onEnd_structure(self, parser, querydict, step=0):
        backend = parser.backend
        self.lastSystemDescription = parser.gidSections["section_system"]
        self.cell = None
        self.labels = None
        celltype = None
        if('basis-atom-coordinates.source-value' in querydict and
           ('a.si-value' in querydict) or
           ('b.si-value' in querydict) or
           ('c.si-value' in querydict) or
           ('a-host.si-value' in querydict) or
           ('b-host.si-value' in querydict) or
           ('c-host.si-value' in querydict)):
            basis = querydict['basis-atom-coordinates.source-value']
            #celltype = ase.geometry.crystal_structure_from_cell(basis)
            if not celltype:
                kimprefix = g(querydict, 'meta.runner.kimid-prefix', None).split('_')
                celltype = kimprefix[1] if len(kimprefix)>1 else None
            if not celltype:
                kimprefix = g(querydict, 'meta.runner.extended-id', None).split('_')
                celltype = kimprefix[1] if len(kimprefix)>1 else None
            if celltype:
                backend.addValue("x_openkim_cubic_crystal_type", celltype)
            if 'species.source-value' in querydict:
                self.labels = querydict['species.source-value']
                species_nums = [atomic_numbers[Z] for Z in self.labels]
                a_si_val = querydict['a.si-value']
                if isinstance(a_si_val, list):
                    lat_a = float(a_si_val[int(step)])
                else:
                    lat_a = float(a_si_val)
                si_conv = convert_unit_function("m", "angstrom")
                cellAtoms = Atoms(
                        positions=basis,
                        cell=[si_conv(lat_a),
                              si_conv(lat_a),
                              si_conv(lat_a)],
                        pbc=True)
                cellAtoms.positions = ase.geometry.wrap_positions(cellAtoms.positions, cellAtoms.cell, pbc=True)
                if len(cellAtoms.numbers) == len(species_nums):
                    cellAtoms.numbers = species_nums
                else:
                    if len(species_nums)<2 and len(cellAtoms.numbers)>1:
                        cellAtoms.numbers = [species_nums[0] for i in cellAtoms.numbers]
                if len(self.labels)<2:
                    self.labels = [self.labels[0] for s in cellAtoms.positions]
                ang_conv = convert_unit_function("angstrom", "m")
                si_cell = np.array([[ang_conv(x) for x in i] for i in cellAtoms.cell])
                si_pos = np.array([[ang_conv(x) for x in i] for i in cellAtoms.positions])
                backend.addArrayValues("simulation_cell", si_cell)
                backend.addArrayValues("configuration_periodic_dimensions", np.ones(3, dtype=bool))
                backend.addArrayValues("atom_positions", si_pos)
        if self.labels is not None:
            backend.addArrayValues("atom_labels", np.asarray(self.labels))

    def onEnd_model(self, parser, kimquery):
        backend = parser.backend
        with open_section(parser, backend, "section_sampling_method"):
            if self.KIM_TE == 0:
                sampling_method = "geometry_optimization"
            else:
                sampling_method = "molecular_dynamics"
            backend.addValue("sampling_method", sampling_method)
        with open_section(parser, backend, "section_frame_sequence"):
            if self.temperature:
                backend.addArrayValues('frame_sequence_temperature_frames', np.array([
                    i for i in self.singleConfCalcs]))
                if isinstance(self.temperature, list):
                    if len(self.temperature)<2:
                        backend.addArrayValues('frame_sequence_temperature', np.array([
                            self.temperature[0] for i in self.singleConfCalcs]))
                    else:
                        backend.addArrayValues('frame_sequence_temperature', np.asarray(self.temperature))
                else:
                    backend.addArrayValues('frame_sequence_temperature', np.asarray([self.temperature]))
            if self.cohesivepot:
                backend.addArrayValues('frame_sequence_potential_energy_frames', np.array([
                    i for i in self.singleConfCalcs]))
                if isinstance(self.cohesivepot, list):
                    if len(self.cohesivepot)<2:
                        backend.addArrayValues('frame_sequence_potential_energy', np.array([
                            self.cohesivepot[0] for i in self.singleConfCalcs]))
                    else:
                        backend.addArrayValues(
                                'frame_sequence_potential_energy', np.asarray(self.cohesivepot))
                else:
                    backend.addArrayValues(
                            'frame_sequence_potential_energy', np.asarray([self.cohesivepot]))
            if self.cohesiveeng:
                backend.addArrayValues('x_openkim_frame_sequence_cohesive_energy_frames', np.array([
                    i for i in self.singleConfCalcs]))
                if isinstance(self.cohesiveeng, list):
                    if len(self.cohesiveeng)<2:
                        backend.addArrayValues('x_openkim_frame_sequence_cohesive_energy', np.array([
                            self.cohesiveeng for i in self.singleConfCalcs]))
                    else:
                        backend.addArrayValues(
                                'x_openkim_frame_sequence_cohesive_energy', np.asarray(self.cohesiveeng))
                else:
                    backend.addArrayValues(
                            'x_openkim_frame_sequence_cohesive_energy', np.asarray([self.cohesiveeng]))
            backend.addArrayValues('frame_sequence_time', np.array([0. for i in self.singleConfCalcs]))
            backend.addValue("frame_sequence_to_sampling_ref", parser.gidSections["section_sampling_method"])
            backend.addArrayValues("frame_sequence_local_frames_ref", np.asarray(self.singleConfCalcs))

    def onEnd_calculation(self, parser, kimquery, step=0):
        backend = parser.backend
        backend.addValue("single_configuration_calculation_to_system_ref", self.lastSystemDescription)
        zeroTemp = None
        if 'temperature.si-value' in kimquery:
            self.temperature = kimquery['temperature.si-value']
        if 'cohesive-potential-energy.si-value' in kimquery:
            self.cohesivepot = kimquery['cohesive-potential-energy.si-value']
            if self.temperature:
                if not isinstance(self.temperature, list):
                    if float(self.temperature) > 0:
                        zeroTemp = False
                    else:
                        zeroTemp = True
            if isinstance(self.cohesivepot, list):
                backend.addValue("energy_total", float(self.cohesivepot[int(step)]))
                backend.addValue("energy_potential", float(self.cohesivepot[int(step)]))
                if zeroTemp:
                    backend.addValue("energy_total_T0", float(self.cohesivepot[int(step)]))
            else:
                backend.addValue("energy_total", float(self.cohesivepot))
                backend.addValue("energy_potential", float(self.cohesivepot))
                if zeroTemp:
                    backend.addValue("energy_total_T0", float(self.cohesivepot))
        if 'cauchy-stress.si-value' in kimquery:
            cstress = kimquery['cauchy-stress.si-value']
            f = np.zeros((3,3))
            f[0][0] = float(cstress[0])
            f[1][1] = float(cstress[1])
            f[2][2] = float(cstress[2])
            f[1][2] = float(cstress[3])
            f[0][2] = float(cstress[4])
            f[0][1] = float(cstress[5])
            backend.addArrayValues("stress_tensor", np.asarray(f))

class KIMParser(object):
    @staticmethod
    def maybeGet(el, meta, default = None):
        if meta in el:
            return el[meta]
        else:
            return default

    def __init__(self, parserInfo, superContext):
        self.fIn = None
        self.parserInfo = parserInfo
        self.superContext = superContext
        self.gidSections = {}

    def parse(self, mainFileUri, fIn, backend):
        self.mainFileUri = mainFileUri
        self.fIn = fIn
        self.backend = backend
        backend.startedParsingSession(
            mainFileUri = mainFileUri,
            parserInfo = self.parserInfo)
        superContext = self.superContext
        superContext.startedParsing(self)
        QueryList = KIMQueryReader(self.fIn)
        try:
            for qi, qdict in enumerate(QueryList):
                superContext.reset()
                with open_section(self, backend, 'section_run'):
                    superContext.onEnd_program(self, qdict)
                    with open_section(self, backend, 'x_openkim_section_metadata'):
                        superContext.onEnd_openkim_data(self, qdict)
                    if 'a.si-value' in qdict or 'a-host.si-value' in qdict:
                        if isinstance(qdict['a.si-value'], list):
                            for step in qdict['a.si-value']:
                                with open_section(self, backend, 'section_single_configuration_calculation'):
                                    superContext.singleConfCalcs.append(self.gidSections[
                                        "section_single_configuration_calculation"])
                                    with open_section(self, backend, 'section_system'):
                                        superContext.onEnd_structure(self, qdict, step)
                                    superContext.onEnd_calculation(self, qdict, step)
                        else:
                            with open_section(self, backend, 'section_single_configuration_calculation'):
                                superContext.singleConfCalcs.append(self.gidSections[
                                    "section_single_configuration_calculation"])
                                with open_section(self, backend, 'section_system'):
                                    superContext.onEnd_structure(self, qdict)
                                superContext.onEnd_calculation(self, qdict)
                    superContext.onEnd_model(self, qdict)
        except:
            logging.exception("failure when parsing %s", self.mainFileUri)
            backend.finishedParsingSession(
                parserStatus = "ParseFailure",
                parserErrors = ["exception: %s" % sys.exc_info()[1]]
            )
        else:
            backend.finishedParsingSession(
                parserStatus = "ParseSuccess",
                parserErrors = None
            )

g = KIMParser.maybeGet

parserInfo = {
  "name": "parser_openkim",
  "version": "1.0"
}


class OpenKIMParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       logging.info('turbomole parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("openkim.nomadmetainfo.json")
       parserInfo = {'name': 'parser_openkim', 'version': '1.0'}
       context = OpenkimContext()
       parser = KIMParser(parserInfo, context)
       parser.parse('nmd://uri', mainfile, backend)

       return backend



from typing import Dict

from optimade.models import Species as OptimadeStructureSpecies
from optimade.models import StructureResource as OptimadeStructure

from optimade.adapters.structures.utils import (
    cell_to_cellpar,
    pad_positions,
    fractional_coordinates,
)

try:
    import numpy as np
except ImportError:
    from warnings import warn

    np = None
    NUMPY_NOT_FOUND = "NumPy not found, cannot convert structure to CIF"


__all__ = ("get_cif",)


def get_cif(  # pylint: disable=too-many-locals,too-many-branches
    optimade_structure: OptimadeStructure,
) -> str:
    """ Get CIF file as string from OPTIMADE structure

    Based on `ase.io.cif:write_cif()`.

    :param optimade_structure: OPTIMADE structure
    :param formatting: What formatting to use for the CIF file data keys.
        Can be either "mp" or "default".
    :param encoding: Encoding used for the string. CIF files use "latin-1" as standard.
        If encoding is "str", a Python str object will be returned.
    :return: str
    """
    # NumPy is needed for calculations
    if globals().get("np", None) is None:
        warn(NUMPY_NOT_FOUND)
        return None

    cif = """#
# Created from an OPTIMADE structure.
#
# See https://www.optimade.org and/or
# https://github.com/Materials-Consortia/OPTIMADE for more information.
#
"""

    cif += f"data_{optimade_structure.id}\n\n"

    attributes = optimade_structure.attributes

    # Do this only if there's three non-zero lattice vectors
    # NOTE: This also negates handling of lattice_vectors with null/None values
    if all(attributes.dimension_types):
        a_vector, b_vector, c_vector, alpha, beta, gamma = cell_to_cellpar(
            attributes.lattice_vectors
        )

        cif += (
            f"_cell_length_a                    {a_vector:g}\n"
            f"_cell_length_b                    {b_vector:g}\n"
            f"_cell_length_c                    {c_vector:g}\n"
            f"_cell_angle_alpha                 {alpha:g}\n"
            f"_cell_angle_beta                  {beta:g}\n"
            f"_cell_angle_gamma                 {gamma:g}\n\n"
        )
        cif += (
            "_symmetry_space_group_name_H-M    'P 1'\n"
            "_symmetry_int_tables_number       1\n\n"
            "loop_\n"
            "  _symmetry_equiv_pos_as_xyz\n"
            "  'x, y, z'\n\n"
        )

        # Since some structure viewers are having issues with cartesian coordinates,
        # we calculate the fractional coordinates if this is a 3D structure and we have all the necessary information.
        if not hasattr(attributes, "fractional_site_positions"):
            sites, _ = pad_positions(attributes.cartesian_site_positions)
            attributes.fractional_site_positions = fractional_coordinates(
                cell=attributes.lattice_vectors, cartesian_positions=sites
            )

    # NOTE: This is otherwise a bit ahead of its time, since this OPTIMADE property is part of an open PR.
    # See https://github.com/Materials-Consortia/OPTIMADE/pull/206
    coord_type = (
        "fract" if hasattr(attributes, "fractional_site_positions") else "Cartn"
    )

    cif += (
        "loop_\n"
        "  _atom_site_type_symbol\n"  # species.chemical_symbols
        "  _atom_site_label\n"  # species.name + unique int
        "  _atom_site_occupancy\n"  # species.concentration
        f"  _atom_site_{coord_type}_x\n"  # cartesian_site_positions
        f"  _atom_site_{coord_type}_y\n"  # cartesian_site_positions
        f"  _atom_site_{coord_type}_z\n"  # cartesian_site_positions
        "  _atom_site_thermal_displace_type\n"  # Set to 'Biso'
        "  _atom_site_B_iso_or_equiv\n"  # Set to 1.0:f
    )

    if coord_type == "fract":
        sites, _ = pad_positions(attributes.fractional_site_positions)
    else:
        sites, _ = pad_positions(attributes.cartesian_site_positions)

    species: Dict[str, OptimadeStructureSpecies] = {
        species.name: species for species in attributes.species
    }

    symbol_occurences = {}
    for site_number in range(attributes.nsites):
        species_name = attributes.species_at_sites[site_number]
        site = sites[site_number]

        current_species = species[species_name]

        for index, symbol in enumerate(current_species.chemical_symbols):
            if symbol == "vacancy":
                continue

            if symbol in symbol_occurences:
                symbol_occurences[symbol] += 1
            else:
                symbol_occurences[symbol] = 1
            label = f"{symbol}{symbol_occurences[symbol]}"

            cif += (
                f"  {symbol} {label} {current_species.concentration[index]:6.4f} {site[0]:8.5f}  "
                f"{site[1]:8.5f}  {site[2]:8.5f}  {'Biso':4}  {'1.000':6}\n"
            )

    return cif



# pylint: disable=no-self-argument,no-name-in-module
import re

from typing import Dict, List, Optional
from pydantic import BaseModel, AnyHttpUrl, Field, validator, root_validator

from .jsonapi import Resource


__all__ = ("AvailableApiVersion", "BaseInfoAttributes", "BaseInfoResource")


class AvailableApiVersion(BaseModel):
    """A JSON object containing information about an available API version"""

    url: AnyHttpUrl = Field(
        ...,
        description="a string specifying a versioned base URL that MUST adhere to the rules in section Base URL",
        pattern=r".+/v[0-1](\.[0-9]+)*/?$",
    )

    version: str = Field(
        ...,
        description="a string containing the full version number of the API served at that versioned base URL. "
        "The version number string MUST NOT be prefixed by, e.g., 'v'.",
        regex=r"^[0-9]+(\.[0-9]+){,2}$",
    )

    @validator("url")
    def url_must_be_versioned_base_url(cls, v):
        """The URL must be a valid versioned Base URL"""
        if not re.match(r".+/v[0-1](\.[0-9]+)*/?$", v):
            raise ValueError(f"url MUST be a versioned base URL. It is: {v}")
        return v

    @root_validator(pre=False, skip_on_failure=True)
    def crosscheck_url_and_version(cls, values):
        """ Check that URL version and API version are compatible. """
        url_version = (
            values["url"]
            .split("/")[-2 if values["url"].endswith("/") else -1]
            .replace("v", "")
        )
        url_version = tuple(int(val) for val in url_version.split("."))
        api_version = tuple(int(val) for val in values["version"].split("."))
        if any(a != b for a, b in zip(url_version, api_version)):
            raise ValueError(
                f"API version {api_version} is not compatible with url version {url_version}."
            )
        return values


class BaseInfoAttributes(BaseModel):
    """Attributes for Base URL Info endpoint"""

    api_version: str = Field(
        ..., description="Presently used version of the OPTIMADE API"
    )
    available_api_versions: List[AvailableApiVersion] = Field(
        ...,
        description="A list of dictionaries of available API versions at other base URLs",
    )
    formats: List[str] = Field(
        default=["json"], description="List of available output formats."
    )
    available_endpoints: List[str] = Field(
        ...,
        description="List of available endpoints (i.e., the string to be appended to the versioned base URL).",
    )
    entry_types_by_format: Dict[str, List[str]] = Field(
        ..., description="Available entry endpoints as a function of output formats."
    )
    is_index: Optional[bool] = Field(
        default=False,
        description="If true, this is an index meta-database base URL (see section Index Meta-Database). "
        "If this member is not provided, the client MUST assume this is not an index meta-database base URL "
        "(i.e., the default is for is_index to be false).",
    )

    @validator("entry_types_by_format", check_fields=False)
    def formats_and_endpoints_must_be_valid(cls, v, values):
        for format_, endpoints in v.items():
            if format_ not in values["formats"]:
                raise ValueError(f"'{format_}' must be listed in formats to be valid")
            for endpoint in endpoints:
                if endpoint not in values["available_endpoints"]:
                    raise ValueError(
                        f"'{endpoint}' must be listed in available_endpoints to be valid"
                    )
        return v


class BaseInfoResource(Resource):
    id: str = Field(default="/", const=True)
    type: str = Field(default="info", const=True)
    attributes: BaseInfoAttributes = Field(...)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module contains functionality to use old 'legacy' NOMAD CoE parsers with the
new nomad@fairdi infrastructure. This covers aspects like the old metainfo, a unifying
wrapper for parsers, parser logging, and a parser backend.
'''

from typing import List, Union, Any, Tuple, Type, cast
from abc import ABCMeta, abstractmethod
import importlib
import os.path
import inspect
from unittest.mock import patch
import logging
import glob
import sys

from nomad import utils, datamodel, config
from nomad.metainfo import (
    SubSection, Quantity, Section, SectionProxy, Reference, MResource, MSection, MSectionBound, Property)
from nomad.metainfo.legacy import (
    LegacyMetainfoEnvironment, python_package_mapping, normalize_name)

from .parser import MatchingParser


class BackendError(Exception):
    pass


class Backend():
    '''
    A backend that uses the new metainfo to store all data.

    Arguments:
        metainfo: The main legacy metainfo package name or a legacy metainfo environment
            instance.
        logger: An optional logger.
        domain: An optional domain name.

    Attributes:
        domain: The domain that this backend contains data for.
        env: The metainfo environment (all available definitions).
        resource: The metainfo resource that contains all data.
        entry_archive: The root section of the archive behind this backend.
        logger: A logger that can be used to log metainfo and backend operation related
            warnings and errors.
    '''

    @staticmethod
    def filter_legacy_defs(definition):
        return definition.m_get_annotations('legacy') is not None

    def __init__(
            self, metainfo: Union[str, LegacyMetainfoEnvironment], domain: str = None,
            entry_archive: datamodel.EntryArchive = None, logger=None):
        assert metainfo is not None

        if logger is None:
            logger = utils.get_logger(__name__)
        self.logger = logger
        self.domain = domain if domain is not None else 'dft'  # TODO

        if isinstance(metainfo, str):
            python_package_name, _ = python_package_mapping(metainfo)
            python_package_name = '.'.join(python_package_name.split('.')[:-1])
            python_module = importlib.import_module(python_package_name)
            metainfo = getattr(python_module, 'm_env')

        self.env: LegacyMetainfoEnvironment = cast(LegacyMetainfoEnvironment, metainfo)
        self.__legacy_env = None
        self.resource = MResource(logger=logger)
        self.entry_archive = datamodel.EntryArchive() if entry_archive is None else entry_archive
        self.resource.add(self.entry_archive)

        self.strict = False  # TODO

        self.reset_status()
        # things that have no real purpose, but are required by some legacy code
        # self._unknown_attributes = {}
        # self._known_attributes = ['results']
        # self.fileOut = io.StringIO()

    def __open_section(self, section_def: Union[Section, SectionProxy], index: int = -1) -> MSection:
        ''' Returns the last opened section with the given parent index. '''
        sections = self.resource.all(section_def.section_cls)

        if sections is None:
            raise KeyError(
                'section %s with given parent index %s does not exist' % (section_def.name, index))

        if index == -1:
            return sections[-1]

        for section in reversed(sections):
            if section.m_parent_index == index:
                return section

        raise KeyError(
            'section %s with given parent index %s does not exist (B)' % (section_def.name, index))

    def __getitem__(self, key):
        property_def = self.resolve_definition(key, Property)
        section_def = property_def.m_parent
        if section_def.extends_base_section:
            section_def = section_def.base_sections[0]

        section = self.__open_section(section_def)
        if isinstance(property_def, Quantity):
            return section.m_get(property_def)

        elif isinstance(property_def, SubSection):
            return section.m_get_sub_sections(property_def)

    def metaInfoEnv(self):
        ''' Returns the meta info used by this backend. '''
        if self.__legacy_env is None:
            self.__legacy_env = self.env.legacy_info_env()
        return self.__legacy_env

    def resolve_definition(self, name, section_cls: Type[MSectionBound]) -> MSectionBound:
        definition = self.env.from_legacy_name(name, section_cls)

        if definition:
            return definition

        return self.env.resolve_definition(
            normalize_name(name), section_cls, Backend.filter_legacy_defs)

    def openSection(self, name, parent_index: int = -1, return_section=False):
        '''
        Opens a new section and returns its new unique gIndex.

        It will assume that there is a sub-section def with the given name.
        It will use the latest opened section of the sub-sections parent as the parent
        for the new section.
        An Exception will be known root sections, e.g. 'section_run'.

        Args:
            name: The section name
            parent_index: Index of the parent section in which the section will
                be opened in.
            return_section: If True, returns the section instead of the section index.
        '''
        section_def = self.resolve_definition(name, Section)

        if name in datamodel.root_sections:
            section = self.entry_archive.m_create(section_def.section_cls)

        else:
            sub_section_def = self.resolve_definition(name, SubSection)
            parent_section_def = sub_section_def.m_parent_as(Section)
            if parent_section_def.extends_base_section:
                parent_section_def = parent_section_def.base_sections[0]

            parent = self.__open_section(parent_section_def, parent_index)
            section = parent.m_create(section_def.section_cls, sub_section_def)

        if return_section:
            return section
        return section.m_parent_index

    def get_open_section_for_quantity(self, name, g_index):
        ''' Returns the open section that contains the quantity of the given name. '''

        quantity_def = self.resolve_definition(name, Quantity)
        section_def = quantity_def.m_parent_as(Section)
        if section_def.extends_base_section:
            section_def = section_def.base_sections[0]

        section = self.__open_section(section_def, g_index)

        return section, quantity_def

    def closeSection(self, name, g_index):
        '''
        Closes the section with the given meta name and index. After this, no more
        value can be added to this section.
        '''
        pass

    def openNonOverlappingSection(self, metaName):
        ''' Opens a new non overlapping section. '''
        return self.openSection(metaName)

    def setSectionInfo(self, metaName, gIndex, references):
        '''
        Sets info values of an open section references should be a dictionary with the
        gIndexes of the root sections this section refers to.
        '''
        # TODO might be necessary to make references work?
        pass

    def closeNonOverlappingSection(self, name):
        '''
        Closes the current non overlapping section for the given meta name. After
        this, no more value can be added to this section.
        '''
        return self.closeSection(name, -1)

    def openSections(self):
        ''' Returns the sections that are still open as metaName, gIndex tuples. '''
        raise NotImplementedError()
        # for section_def, sub_sections in self.open_sections_by_def:
        #     for sub_section in sub_sections:
        #         yield section_def.name, sub_section.m_parent_index

    def addValue(self, name, value, g_index=-1):
        '''
        Adds a json value for the given metaName. The gIndex is used to identify
        the right parent section.
        '''
        section, quantity_def = self.get_open_section_for_quantity(name, g_index)
        if isinstance(quantity_def.type, Reference):
            # quantity is a reference
            try:
                referenced_target = self.__open_section(
                    quantity_def.type.target_section_def, value)
            except KeyError as e:
                raise BackendError(str(e))

            value = referenced_target

        setattr(section, name, value)

    def addRealValue(self, name, value, g_index=-1):
        '''
        Adds a float value for the given metaName. The gIndex is used to identify
        the right parent section.
        '''
        self.addValue(name, value, g_index)

    def addArray(self, name, shape, g_index=-1):
        '''
        Adds an uninitialized array of the given shape for the given metaName.
        The gIndex is used to identify the right parent section.
        This is neccessary before array values can be set with :func:`setArrayValues`.
        '''
        raise NotImplementedError()

    def setArrayValues(self, metaName, values, offset=None, gIndex=-1):
        '''
        Adds values of the given numpy array to the last array added for the given
        metaName and parent gIndex.
        '''
        raise NotImplementedError()

    def addArrayValues(self, name, values, gIndex=-1, override: bool = False):
        '''
        Adds an array with the given numpy array values for the given metaName and
        parent section gIndex. Override determines whether to rewrite exisiting values
        in the backend.
        '''
        section, quantity_def = self.get_open_section_for_quantity(name, gIndex)
        if isinstance(quantity_def.type, Reference):
            # quantity is a reference
            resolved_values = []
            for value in values:
                # quantity is a reference
                try:
                    referenced_target = self.__open_section(quantity_def.type.target_section_def, value)
                except KeyError:
                    raise BackendError('There is not section for the given reference index')

                resolved_values.append(referenced_target)
            values = resolved_values

        if self.strict and not override:
            quantity_def = section.m_def.all_quantities[name]
            assert not section.m_is_set(quantity_def)

        setattr(section, name, values)

    # The following are extensions to the origin NOMAD-coe parser backend. And allow
    # access to existing data

    def get_sections(self, meta_name: str, g_index: int = -1) -> List[int]:
        ''' Return all gIndices for existing sections of the given meta_name and parent index. '''
        section_def = self.resolve_definition(meta_name, Section)
        return [
            section.m_parent_index for section in self.resource.all(section_def.section_cls)
            if g_index == -1 or section.m_parent.m_parent_index == g_index]

    def get_value(self, meta_name: str, g_index=-1) -> Any:
        '''
        Return the value set to the given meta_name in its parent section of the given index.
        An index of -1 (default) is only allowed if there is exactly one parent section.
        '''
        section, quantity_def = self.get_open_section_for_quantity(meta_name, g_index)
        value = section.m_get(quantity_def)

        if value is None:
            raise KeyError(quantity_def.name)

        if isinstance(quantity_def.type, Reference):
            return value.m_parent_index

        return value

    def startedParsingSession(
            self, mainFileUri, parserInfo, parserStatus=None, parserErrors=None):
        '''
        Should be called when the parsing starts.
        ParserInfo should be a valid json dictionary.
        '''
        self.reset_status()

    def finishedParsingSession(
            self, parserStatus, parserErrors, mainFileUri=None, parserInfo=None,
            parsingStats=None):
        ''' Called when the parsing finishes. '''
        self._status = parserStatus
        self._errors = parserErrors

    def addMatchTelemetry(self, match_telemetry, gIndex=-1):
        # TODO
        pass

    def pwarn(self, msg):
        ''' Used to catch parser warnings. '''
        self.logger.warn(msg)
        if len(self._warnings) < 10:
            self._warnings.append(msg)
        elif len(self._warnings) == 10:
            self._warnings.append('There are more warnings, check the processing logs.')

    @property
    def status(self) -> Tuple[str, List[str]]:
        ''' Returns status and potential errors. '''
        return (self._status, self._errors)

    def reset_status(self) -> None:
        self._status = 'ParseSuccess'
        self._errors = None
        self._warnings: List[str] = []


class LegacyParser(MatchingParser):
    '''
    A parser implementation for legacy NOMAD-coe parsers. It assumes that parsers
    are installed to the python environment.

    Arguments:
        parser_class_name: the main parser class that implements NOMAD-coe's
        backend_factory: a callable that returns a backend, takes meta_info and logger as argument
    '''
    def __init__(self, parser_class_name: str, *args, backend_factory=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.parser_class_name = parser_class_name
        self.backend_factory = backend_factory

        module_name = self.parser_class_name.split('.')[:-1]
        parser_class_name = self.parser_class_name.split('.')[-1]
        self.__parser_impl = module_name, parser_class_name
        self.__parser_class = None

    @property
    def metainfo_env(self):
        if self._metainfo_env is None:
            module_name, _ = self.__parser_impl
            module = importlib.import_module('.'.join(module_name + ['metainfo']))
            self._metainfo_env = getattr(module, 'm_env')

        return super().metainfo_env

    @property
    def parser_class(self):
        if self.__parser_class is None:
            module_name, parser_class_name = self.__parser_impl
            module = importlib.import_module('.'.join(module_name))
            self.__parser_class = getattr(module, parser_class_name)

        return self.__parser_class

    def parse(self, mainfile: str, archive: datamodel.EntryArchive, logger=None):
        # TODO we need a homogeneous interface to parsers, but we dont have it right now.
        # There are some hacks to distinguish between ParserInterface parser and simple_parser
        # using hasattr, kwargs, etc.
        if logger is None:
            logger = utils.get_logger('__name__')

        if issubclass(self.parser_class, CoEParser):
            # TODO reuse parser
            # TODO remove, this whole mechanism is only used by wien2k
            parser = self.parser_class()  # pylint: disable=not-callable
            backend = parser.run(mainfile, logger=logger, entry_archive=archive)

        else:
            def create_backend(meta_info):
                if self.backend_factory is not None:
                    return self.backend_factory(meta_info, logger=logger)

                return Backend(meta_info, logger=logger, domain=self.domain, entry_archive=archive)

            init_signature = inspect.getargspec(self.parser_class.__init__)
            kwargs = dict(backend=create_backend, log_level=logging.DEBUG, debug=True)
            kwargs = {key: value for key, value in kwargs.items() if key in init_signature.args}

            with utils.legacy_logger(logger):
                self.parser = self.parser_class(**kwargs)  # pylint: disable=not-callable

                with patch.object(sys, 'argv', []):
                    backend = self.parser.parse(mainfile)

        os.chdir(config.fs.working_directory)

        if backend is None or not hasattr(backend, 'status'):
            backend = self.parser.parser_context.super_backend

        if backend.status[0] != 'ParseSuccess':
            raise Exception(backend.status[1])


class CoEParser(metaclass=ABCMeta):

    @abstractmethod
    def run(self, mainfile, logger) -> Backend:
        pass


class CoEInterfaceParser(CoEParser):

    def __init__(self, interface_class):
        super().__init__()
        self.interface_class = interface_class
        self.__interface = None

    def run(self, mainfile, logger):
        if self.__interface is None:
            self.__interface = self.interface_class()

        self.__interface.setup_logger(logger)
        self.__interface.parse(mainfile)
        return self.__interface.parser_context.super_backend


class CoESimpleMatcherParser(CoEParser):

    def __init__(self):
        super().__init__()
        self.parser_description = self.create_parser_description()
        self.simple_matcher = self.create_simple_matcher()

        self._metainfo_env = self.metainfo_env()
        self.__legacy_metainfo_env = None

        self.caching_levels = self.create_caching_levels()

    @abstractmethod
    def metainfo_env(self) -> LegacyMetainfoEnvironment:
        pass

    @property
    def metaInfoEnv(self) -> LegacyMetainfoEnvironment:
        if self.__legacy_metainfo_env is None:
            self.__legacy_metainfo_env = self._metainfo_env.legacy_info_env()
        return self.__legacy_metainfo_env

    def create_caching_levels(self) -> dict:
        return dict()

    @abstractmethod
    def create_simple_matcher(self):
        pass

    @abstractmethod
    def create_parser_description(self) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def create_super_context(self):
        pass

    def simple_parser(self, mainfile, **kwargs) -> Backend:
        from nomadcore.simple_parser import mainFunction
        backend = Backend(self._metainfo_env, **kwargs)
        from unittest.mock import patch
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=self.simple_matcher,
                metaInfoEnv=self.metaInfoEnv,
                parserInfo=self.parser_description,
                cachingLevelForMetaName=self.caching_levels,
                superContext=self.create_super_context(),
                superBackend=backend)

        return backend

    def run(self, mainfile, logger, **kwargs) -> Backend:
        with utils.legacy_logger(logger):
            return self.simple_parser(mainfile, logger=logger, **kwargs)


class VaspOutcarParser(LegacyParser):
    '''
    LegacyParser that only matches mailfiles, if there is no .xml in the
    same directory, i.e. to use the VASP OUTCAR parser in absence of .xml
    output file.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = 'parsers/vaspoutcar'

    def is_mainfile(self, filename: str, *args, **kwargs) -> bool:
        is_mainfile = super().is_mainfile(filename, *args, **kwargs)

        if is_mainfile:
            directory = os.path.dirname(filename)
            if len(glob.glob('%s/*.xml*' % directory)) > 0:
                return False

        return is_mainfile



# Copyright 2016-2018 Fawzi Mohamed, Lauri Himanen, Danio Brambila, Ankit Kariryaa, Henning Glawe
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
from builtins import range
from builtins import object
import xml.etree.ElementTree
from xml.etree.ElementTree import ParseError
import bisect
from datetime import datetime
import re
import traceback
import numpy as np
import ase.geometry
import ase.data
import xml.etree.ElementTree as ET
import logging

from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit

from nomad.metainfo import Quantity
from nomad.datamodel.metainfo.public import section_run as msection_run
from nomad.datamodel.metainfo.public import section_method as msection_method
from nomad.datamodel.metainfo.public import section_sampling_method as msection_sampling_method
from nomad.datamodel.metainfo.public import section_frame_sequence as msection_frame_sequence
from nomad.datamodel.metainfo.public import section_single_configuration_calculation as msection_single_configuration_calculation
from nomad.datamodel.metainfo.public import section_basis_set as msection_basis_set
from nomad.datamodel.metainfo.public import section_XC_functionals as msection_XC_functionals
from nomad.datamodel.metainfo.public import section_system as msection_system
from nomad.datamodel.metainfo.public import section_k_band as msection_k_band
from nomad.datamodel.metainfo.public import section_k_band_segment as msection_k_band_segment
from nomad.datamodel.metainfo.public import section_k_band_normalized as msection_k_band_normalized
from nomad.datamodel.metainfo.public import section_k_band_segment_normalized as msection_k_band_segment_normalized
from nomad.datamodel.metainfo.public import section_eigenvalues as msection_eigenvalues
from nomad.datamodel.metainfo.public import section_method_atom_kind as msection_method_atom_kind
from nomad.datamodel.metainfo.public import section_basis_set_cell_dependent as msection_basis_set_cell_dependent
from nomad.datamodel.metainfo.public import section_dos as msection_dos
from nomad.datamodel.metainfo.common import section_method_basis_set as msection_method_basis_set

eV2J = convert_unit_function("eV", "J")
eV2JV = np.vectorize(eV2J)

vasp_to_metainfo_type_mapping = {
    'string': ['C'],
    'int': ['i'],
    'logical': ['b', 'C'],
    'float': ['f']}

special_paths = {
    'cubic': 'ΓXMΓRX,MR',
    'fcc': 'ΓXWKΓLUWLK,UX',
    'bcc': 'ΓHNΓPH,PN',
    'tetragonal': 'ΓXMΓZRAZXR,MA',
    'orthorhombic': 'ΓXSYΓZURTZ,YT,UX,SR',
    'hexagonal': 'ΓMKΓALHA,LM,KH',
    'monoclinic': 'ΓYHCEM1AXH1,MDZ,YD'}


def seconds_from_epoch(date):
    epoch = datetime(1970, 1, 1)
    ts = date - epoch
    return ts.seconds + ts.microseconds / 1000.0


trueRe = re.compile(
    r"\s*(?:\.?[Tt](?:[Rr][Uu][Ee])?\.?|1|[Yy](?:[Ee][Ss])?|[Jj][Aa]?)\s*$")
falseRe = re.compile(
    r"\s*(?:\.?[fF](?:[Aa][Ll][Ss][Ee])?\.?|0|[Nn](?:[Oo]|[Ee][Ii][Nn])?)\s*$")


def to_bool(value):
    if falseRe.match(value):
        return False
    elif trueRe.match(value):
        return True
    else:
        return None


meta_type_transformers = {
    'C': lambda x: x.strip(),
    'i': lambda x: int(float(x.strip())),
    'f': lambda x: float(x.strip()),
    'b': to_bool,
}


class MyXMLParser(ET.XMLParser):

    rx = re.compile("&#([0-9]+);|&#x([0-9a-fA-F]+);")

    def feed(self, data):
        m = self.rx.search(data)
        if m is not None:
            target = m.group(1)
            if target:
                num = int(target)
            else:
                num = int(m.group(2), 16)
            if not(
                num in (0x9, 0xA, 0xD) or 0x20 <= num <= 0xD7FF or 0xE000 <= num <= 0xFFFD
                or 0x10000 <= num <= 0x10FFFF):
                # is invalid xml character, cut it out of the stream
                mstart, mend = m.span()
                mydata = data[:mstart] + data[mend:]
        else:
            mydata = data
        super(MyXMLParser, self).feed(mydata)


def transform2(y):
    if '**' in y:
        return float('nan')
    else:
        return y


def get_vector(el, transform=float, field="v"):
    """ returns the vasp style vector contained in the element el (using field v).
    single elements are converted using the function convert"""
#
#    for x in el.findall(field):
#        for y in re.split(r"\s+", x.text.strip()):
    return [[transform(transform2(y)) for y in re.split(r"\s+", x.text.strip())] for x in el.findall(field)]


class VasprunContext(object):

    def __init__(self, logger=None):
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger

        self._parser = None
        self.bands = None
        self.kpoints = None
        self.weights = None
        self.tetrahedrons = None
        self.tetrahedronVolume = None
        self.ispin = None
        self.ibrion = None
        self.labels = None
        self.vbTopE = None
        self.ebMinE = None
        self.e_fermi = None
        self.cell = None
        self.angstrom_cell = None
        self.unknown_incars = {}
        self._energies = []

    section_map = {
        "modeling": ["section_run", "section_method"],
        "structure": ["section_system"],
        "calculation": ["section_single_configuration_calculation"]
    }

    @property
    def parser(self):
        return self._parser

    @parser.setter
    def parser(self, name):
        self._parser = name

    def on_end_generator(self, element, path_str):
        run = self.parser.run
        program_name = g(element, "i/[@name='program']")
        if program_name.strip().upper() == 'VASP':
            run.program_name = 'VASP'
        else:
            raise Exception('unexpected program name: %s' % program_name)

        version = (
            g(element, "i/[@name='version']", "") + " " +
            g(element, "i/[@name='subversion']", "") + " " +
            g(element, "i/[@name='platform']", ""))

        if not version.isspace():
            run.program_version = version

        run.program_basis_set_type = "plane waves"
        date = g(element, "i/[@name='date']")
        pdate = None
        time = g(element, "i/[@name='time']")

        if date:
            pdate = datetime.strptime(date.strip(), "%Y %m %d")
        if pdate and time:
            pdate = datetime.combine(pdate.date(), datetime.strptime(
                time.strip(), "%H:%M:%S").timetz())
        if pdate:
            run.program_compilation_datetime = seconds_from_epoch(pdate)
        for i in element:
            if i.tag != "i" or not i.attrib.get("name") in set(
                [
                    'program', 'version', 'subversion', 'platform', 'program_version',
                    'date', 'time']):
                pass

    def on_end_incar(self, element, path_str):
        run = self.parser.run
        m_env = self.parser.metainfo_env
        section_method = run.m_create(msection_method)
        dft_plus_u = False
        ibrion = None
        nsw = 0
        for el in element:
            if el.tag == "v":
                name = el.attrib.get("name", None)
                meta = None
                try:
                    meta = m_env.resolve_definition('x_vasp_incar_' + name, Quantity)
                except KeyError:
                    pass
                if not meta:
                    self.logger.warn(
                        "Unknown INCAR parameter (not registered in the meta data): %s %s %r" % (
                            el.tag, el.attrib, el.text))
                    continue
                vector_val = np.asarray(get_vector(el))
                setattr(section_method, meta.get('name'), vector_val)
            elif el.tag == "i":
                name = el.attrib.get("name", None)
                meta = None
                try:
                    meta = m_env.resolve_definition('x_vasp_incar_' + name, Quantity)
                except KeyError:
                    pass
                val_type = el.attrib.get("type")
                if not meta:
                    self.logger.warn(
                        "Unknown INCAR parameter (not registered in the meta data): %s %s %r" % (
                            el.tag, el.attrib, el.text))
                elif val_type:
                    expected_meta_type = {
                        'string': ['C'],
                        'int': ['i'],
                        'logical': ['b', 'C']
                    }.get(val_type)
                    metainfo_type = self._metainfo_type(meta)
                    if not expected_meta_type:
                        self.logger.warn("Unknown value type %s encountered in INCAR: %s %s %r" % (
                            val_type, el.tag, el.attrib, el.text))
                    elif not metainfo_type in expected_meta_type:
                        self.logger.warn("type mismatch between meta data %s and INCAR type %s for %s %s %r" % (
                            metainfo_type, val_type, el.tag, el.attrib, el.text))
                    else:
                        shape = meta.get("shape")
                        converter = meta_type_transformers.get(metainfo_type)
                        if not converter:
                            self.logger.warn(
                                "could not find converter for %s when handling meta info %s" % (
                                    metainfo_type, meta))
                        elif shape:
                            vals = re.split(r"\s+", el.text.strip())
                            setattr(section_method, meta['name'], [converter(x) for x in vals])
                        else:
                            setattr(section_method, meta["name"], converter(el.text))
                    if name == 'GGA':
                        # FIXME tmk: many options are not coded yet. See
                        # https://www.vasp.at/wiki/index.php/GGA
                        f_map = {
                            '91': ['GGA_X_PW91', 'GGA_C_PW91'],
                            'PE': ['GGA_X_PBE', 'GGA_C_PBE'],
                            'RP': ['GGA_X_RPBE', 'GGA_C_PBE'],
                            'PS': ['GGA_C_PBE_SOL', 'GGA_X_PBE_SOL'],
                            'MK': ['GGA_X_OPTB86_VDW']
                        }
                        functs = f_map.get(el.text.strip(), None)
                        if not functs:
                            self.logger.warn("Unknown XC functional %s" % el.text.strip())
                        else:
                            for f in functs:
                                section_XC_functionals = section_method.m_create(msection_XC_functionals)
                                section_XC_functionals.XC_functional_name = f
                    elif name == "ISPIN":
                        self.ispin = int(el.text.strip())
                    elif name == "LDAU":
                        if re.match(".?[Tt](?:[rR][uU][eE])?.?|[yY](?:[eE][sS])?|1", el.text.strip()):
                            dft_plus_u = True
                    elif name == "IBRION":
                        ibrion = int(el.text.strip())
                    elif name == "NSW":
                        nsw = int(el.text.strip())
            else:
                self.logger.warn(
                    "unexpected tag %s %s %r in incar" % (el.tag, el.attrib, el.text))

        if ibrion is None:
            ibrion = -1 if nsw == 0 or nsw == 1 else 0
        if nsw == 0:
            ibrion = -1
        self.ibrion = ibrion
        section_method.electronic_structure_method = 'DFT+U' if dft_plus_u else 'DFT'

    def on_end_kpoints(self, element, path_str):
        run = self.parser.run
        self.bands = None
        self.kpoints = None
        self.weights = None
        for el in element:
            if el.tag == "generation":
                param = el.attrib.get("param", None)  # eg. listgenerated, Monkhorst-Pack, Gamma
                if param:
                    run.section_method[-1].x_vasp_k_points_generation_method = param
                if param == "listgenerated":
                    # This implies a path on k-space, potentially a bandstructure calculation
                    # Save k-path info into a dictionary
                    self.bands = {
                        "divisions": g(el, "i/[@name='divisions']", None),
                        "points": get_vector(el)
                    }

                elif param in ["Monkhorst-Pack", "Gamma"]:
                    # This implies a (2D|3D) mesh on k-space, i.e., not a badstructure calculation
                    # Hence, do nothing: k-points will be stored in the `varray` if-block
                    pass
                else:
                    self.logger.warn("Unknown k point generation method '%s'" % param)
            elif el.tag == "varray":
                name = el.attrib.get("name", None)
                if name == "kpointlist":
                    self.kpoints = np.asarray(get_vector(el))
                    run.section_method[-1].k_mesh_points = self.kpoints
                elif name == "weights":
                    self.weights = np.asarray(get_vector(el))
                    run.section_method[-1].k_mesh_weights = self.weights.flatten()
                elif name == "tetrahedronlist":
                    self.tetrahedrons = np.asarray(get_vector(el), dtype=np.int)
                    run.section_method[-1].x_vasp_tetrahedrons_list = self.tetrahedrons
                else:
                    self.logger.warn("Unknown array %s in kpoints" % name)
            elif el.tag == "i":
                name = el.attrib.get("name", None)
                if name == "volumeweight":
                    ang2m = convert_unit_function("angstrom", "m")

                    # get volume and transform to meters^3
                    vol_cubic_angs = float(el.text.strip())
                    vol_cubic_meters = ang2m(ang2m(ang2m(vol_cubic_angs)))

                    run.section_method[-1].x_vasp_tetrahedron_volume = vol_cubic_meters
            else:
                self.logger.pwarn("Unknown tag %s in kpoints" % el.tag)

    def on_end_structure(self, element, path_str):
        run = self.parser.run
        section_system = run.m_create(msection_system)
        self.cell = None
        for el in element:
            if (el.tag == "crystal"):
                for cell_el in el:
                    if cell_el.tag == "varray":
                        name = cell_el.attrib.get("name", None)
                        if name == "basis":
                            conv = convert_unit_function("angstrom", "m")
                            self.cell = get_vector(
                                cell_el, lambda x: conv(float(x)))
                            self.angstrom_cell = np.array(get_vector(cell_el))
                            section_system.simulation_cell = np.asarray(self.cell)
                            section_system.configuration_periodic_dimensions = np.ones(3, dtype=bool)
                        elif name == "rec_basis":
                            pass
                        else:
                            self.logger.warn(
                                "Unexpected varray %s in crystal" % name)
                    elif cell_el.tag == "i":
                        if cell_el.attrib.get("name") != "volume":
                            self.logger.warn(
                                "Unexpected i value %s in crystal" % cell_el.attrib)
                    else:
                        self.logger.warn("Unexpected tag %s %s %r in crystal" % (
                            cell_el.tag, cell_el.attrib, cell_el.text))
            elif el.tag == "varray":
                name = el.attrib.get("name", None)
                if name == "positions":
                    pos = get_vector(el)
                    section_system.atom_positions = np.dot(np.asarray(pos), self.cell)
                elif name == "selective":
                    atom_sel = get_vector(el, transform=lambda item: item == 'T')
                    section_system.x_vasp_selective_dynamics = np.asarray(atom_sel, dtype=np.bool)
                else:
                    self.logger.warn(
                        "Unexpected varray in structure %s" % el.attrib)
            elif el.tag == "nose":
                nose = get_vector(el)
                section_system.x_vasp_nose_thermostat = nose
            else:
                self.logger.warn(
                    "Unexpected tag in structure %s %s %r" % (el.tag, el.attrib, el.text))
        if self.labels is not None:
            section_system.atom_labels = self.labels

    def on_end_eigenvalues(self, element, path_str):
        if path_str != "modeling/calculation/eigenvalues":
            return True
        run = self.parser.run
        eigenvalues = None
        occupation = None
        for el in element:
            if el.tag == "array":
                for arrray_el in el:
                    if arrray_el.tag == "dimension":
                        pass
                    elif arrray_el.tag == "field":
                        pass
                    elif arrray_el.tag == "set":
                        isp = -1
                        for spin_el in arrray_el:
                            if spin_el.tag == "set":
                                ik = -1
                                isp += 1
                                for kEl in spin_el:
                                    if kEl.tag == "set":
                                        ik += 1
                                        bands = np.asarray(
                                            get_vector(kEl, field="r"))
                                        if eigenvalues is None:
                                            eigenvalues = np.zeros(
                                                (self.ispin, self.kpoints.shape[0], bands.shape[0]), dtype=float)
                                            occupation = np.zeros(
                                                (self.ispin, self.kpoints.shape[0], bands.shape[0]), dtype=float)
                                        eigenvalues[isp, ik] = bands[:, 0]
                                        occupation[isp, ik] = bands[:, 1]
                                    else:
                                        self.logger.warn(
                                            "unexpected tag %s in k array of the eigenvalues" % kEl.tag)
                            else:
                                self.logger.warn(
                                    "unexpected tag %s in spin array of the eigenvalues" % spin_el.tag)
                    else:
                        self.logger.warn(
                            "unexpected tag %s in array of the eigenvalues" % arrray_el.tag)
                if eigenvalues is not None:

                    ev = eV2JV(eigenvalues)
                    vbTopE = []
                    ebMinE = []
                    for ispin in range(occupation.shape[0]):
                        vbTopE.append(float('-inf'))
                        ebMinE.append(float('inf'))
                        for ik in range(occupation.shape[1]):
                            ebIndex = bisect.bisect_right(
                                -occupation[ispin, ik, :], -0.5) - 1
                            vbTopIndex = ebIndex - 1
                            if vbTopIndex >= 0:
                                vbTopK = ev[ispin, ik, vbTopIndex]
                                if vbTopK > vbTopE[ispin]:
                                    vbTopE[ispin] = vbTopK
                            if ebIndex < ev.shape[2]:
                                ebMinK = ev[ispin, ik, ebIndex]
                                if ebMinK < ebMinE[ispin]:
                                    ebMinE[ispin] = ebMinK
                    self.vbTopE = vbTopE
                    self.ebMinE = ebMinE
                    run.section_single_configuration_calculation[-1].energy_reference_highest_occupied = np.array(vbTopE)
                    run.section_single_configuration_calculation[-1].energy_reference_lowest_unoccupied = np.array(ebMinE)
                    if self.bands:
                        divisions = int(self.bands['divisions'])
                        section_k_band = run.section_single_configuration_calculation[-1].m_create(msection_k_band)
                        nsegments = self.kpoints.shape[0] // divisions
                        kpt = np.reshape(
                            self.kpoints, (nsegments, divisions, 3))
                        energies = np.reshape(
                            ev, (self.ispin, nsegments, divisions, bands.shape[0]))
                        occ = np.reshape(
                            occupation, (self.ispin, nsegments, divisions, bands.shape[0]))
                        for isegment in range(nsegments):
                            section_k_band_segment = section_k_band.m_create(msection_k_band_segment)
                            section_k_band_segment.band_energies = energies[:, isegment, :, :]
                            section_k_band_segment.band_occupations = occ[:, isegment, :, :]
                            section_k_band_segment.band_k_points = kpt[isegment]
                            # "band_segm_labels"
                            section_k_band_segment.band_segm_start_end = np.asarray(
                                [kpt[isegment, 0], kpt[isegment, divisions - 1]])

                        section_k_band_normalized = run.section_single_configuration_calculation[-1].m_create(msection_k_band_normalized)
                        for isegment in range(nsegments):
                            section_k_band_segment_normalized = section_k_band_normalized.m_create(msection_k_band_segment_normalized)
                            section_k_band_segment_normalized.band_energies_normalized = energies[:, isegment, :, :] - max(self.vbTopE)
                            section_k_band_segment_normalized.band_occupations_normalized = occ[:, isegment, :, :]
                            section_k_band_segment_normalized.band_k_points_normalized = kpt[isegment]
                            section_k_band_segment_normalized.band_segm_start_end_normalized = np.asarray(
                                [kpt[isegment, 0], kpt[isegment, divisions - 1]])

                    else:
                        section_eigenvalues = run.section_single_configuration_calculation[-1].m_create(msection_eigenvalues)
                        section_eigenvalues.eigenvalues_values = ev
                        section_eigenvalues.eigenvalues_occupation = occupation
            else:
                self.logger.warn("unexpected tag %s in the eigenvalues" % el.tag)

    def on_start_calculation(self, element, path_str):
        run = self.parser.run
        sscc = run.m_create(msection_single_configuration_calculation)
        if self.waveCut:
            section_basis_set = sscc.m_create(msection_basis_set)
            section_basis_set.mapping_section_basis_set_cell_dependent = self.waveCut

    def on_end_modeling(self, element, path_str):
        run = self.parser.run
        run.section_method[-1].x_vasp_unknown_incars = self.unknown_incars
        if self.ibrion is None or self.ibrion == -1:
            return
        section_sampling_method = run.m_create(msection_sampling_method)
        if self.ibrion == 0:
            sampling_method = "molecular_dynamics"
        else:
            sampling_method = "geometry_optimization"
        section_sampling_method.sampling_method = sampling_method
        section_frame_sequence = run.m_create(msection_frame_sequence)
        section_frame_sequence.frame_sequence_to_sampling_ref = section_sampling_method
        section_frame_sequence.frame_sequence_local_frames_ref = run.section_single_configuration_calculation[-1]

    def on_end_calculation(self, element, path_str):
        e_conv = eV2J
        f_conv = convert_unit_function("eV/angstrom", "N")
        p_conv = convert_unit_function("eV/angstrom^3", "Pa")
        run = self.parser.run
        sscc = run.section_single_configuration_calculation[-1]
        sscc.single_configuration_calculation_to_system_ref = run.section_system[-1]
        sscc.single_configuration_to_calculation_method_ref = run.section_method[-1]
        for el in element:
            if el.tag == "energy":
                for en_el in el:
                    if en_el.tag == "i":
                        name = en_el.attrib.get("name", None)
                        if name == "e_fr_energy":
                            value = e_conv(float(en_el.text.strip()))
                            sscc.energy_free = value
                        elif name == "e_wo_entrp":
                            value = e_conv(float(en_el.text.strip()))
                            sscc.energy_total = value
                        elif name == "e_0_energy":
                            value = e_conv(float(en_el.text.strip()))
                            sscc.energy_total_T0 = value
                            self._energies.append(value)
                    elif en_el.tag == "varray":
                        pass

            elif el.tag == 'varray':
                name = el.attrib.get("name", None)
                if name == "forces":
                    f = get_vector(el, lambda x: f_conv(float(x)))
                    sscc.atom_forces = f
                elif name == 'stress':
                    f = get_vector(el, lambda x: p_conv(float(x)))
                    sscc.stress_tensor = f

    def on_end_atominfo(self, element, path_str):
        run = self.parser.run
        atom_types = []
        labels = []
        labels2 = None
        atom_types_desc = []
        for el in element:
            if el.tag in ['atoms', 'types']:
                pass

            elif el.tag == "array":
                name = el.attrib.get("name", None)
                if name == "atoms":
                    for atoms_el in el:
                        if atoms_el.tag == "dimension":
                            pass
                        elif atoms_el.tag == "field":
                            pass
                        elif atoms_el.tag == "set":
                            for atoms_line in atoms_el:
                                if atoms_line.tag != "rc":
                                    self.logger.warn(
                                        "unexpected tag %s in atoms array in atominfo" % atoms_line.tag)
                                else:
                                    line = atoms_line.findall("c")
                                    labels.append(line[0].text.strip())
                                    atom_types.append(int(line[1].text.strip()))
                        else:
                            self.logger.warn(
                                "unexpected tag %s in atoms array in atominfo" % atoms_el.tag)
                elif name == "atomtypes":
                    keys = []
                    field_types = []
                    for atoms_el in el:
                        if atoms_el.tag == "dimension":
                            pass
                        elif atoms_el.tag == "field":
                            keys.append(atoms_el.text.strip())
                            field_types.append(
                                atoms_el.attrib.get("type", "float"))
                        elif atoms_el.tag == "set":
                            expected_keys = [
                                "atomspertype", "element", "mass", "valence", "pseudopotential"]
                            if keys != expected_keys:
                                self.logger.warn(
                                    "unexpected fields in atomtype: %s vs %s" % (keys, expected_keys))
                            for atoms_line in atoms_el:
                                if atoms_line.tag != "rc":
                                    self.logger.warn(
                                        "unexpected tag %s in atoms array in atominfo" % atoms_line.tag)
                                else:
                                    line = atoms_line.findall("c")
                                    type_desc = {}
                                    for i, k in enumerate(keys):
                                        field_type = field_types[i]
                                        value = line[i].text
                                        if field_type == "float":
                                            value = float(value)
                                        elif field_type == "int":
                                            value = int(value)
                                        else:
                                            pass
                                        type_desc[k] = value
                                    atom_types_desc.append(type_desc)
                        else:
                            self.logger.warn(
                                "unexpected tag %s in atomtypes array in atominfo" % atoms_el.tag)
                    n_el = {}
                    kind_labels = []
                    for atom_desc in atom_types_desc:
                        section_method_atom_kind = run.section_method[-1].m_create(msection_method_atom_kind)
                        if 'element' in atom_desc:
                            elName = atom_desc['element'].strip()
                            try:
                                elNr = ase.data.chemical_symbols.index(elName)
                                section_method_atom_kind.method_atom_kind_atom_number = elNr
                            except Exception as e:
                                self.logger.error(
                                    "error finding element number for %r" % atom_desc['element'].strip(),
                                    exc_info=e)
                            n_el_now = 1 + n_el.get(elName, 0)
                            n_el[elName] = n_el_now
                            el_label = elName + \
                                (str(n_el_now) if n_el_now > 1 else "")
                            kind_labels.append(el_label)
                            section_method_atom_kind.method_atom_kind_label = el_label
                            if "mass" in atom_desc:
                                section_method_atom_kind.method_atom_kind_mass = atom_desc["mass"]
                            if "valence" in atom_desc:
                                section_method_atom_kind.method_atom_kind_explicit_electrons = atom_desc["valence"]
                            if "pseudopotential" in atom_desc:
                                section_method_atom_kind.method_atom_kind_pseudopotential_name = atom_desc["pseudopotential"]
                    run.section_method[-1].x_vasp_atom_kind_refs = run.section_method[-1].section_method_atom_kind
                    labels2 = [kind_labels[i - 1] for i in atom_types]
                else:
                    self.logger.warn(
                        "unexpected array named %s in atominfo" % name)
            else:
                self.logger.warn("unexpected tag %s in atominfo" % el.tag)
        self.labels = np.asarray(labels2) if labels2 else np.asarray(labels)

    def _metainfo_type(self, meta):
        dtype = meta.type
        if dtype == str:
            return 'C'
        elif dtype == bool:
            return 'b'
        elif dtype == int or type(dtype) == np.dtype:
            return 'i'
        elif dtype == float:
            return 'f'
        else:
            return

    def _incar_out_tag(self, el):
        if (el.tag != "i"):
            self.logger.warn("unexpected tag %s %s %r in incar" % (el.tag, el.attrib, el.text))

        else:
            run = self.parser.run
            name = el.attrib.get("name", None)
            val_type = el.attrib.get("type")
            try:
                meta = self.parser.metainfo_env.resolve_definition('x_vasp_incarOut_' + name, Quantity)
            except Exception:
                return

            if not meta:
                # Unknown_Incars_Begin: storage into a dictionary
                if not val_type:
                    # On vasp's xml files, val_type *could* be absent if incar value is float
                    val_type = 'float'

                # map vasp's datatype to nomad's datatype [b, f, i, C, D, R]
                nomad_dtype_str = vasp_to_metainfo_type_mapping[val_type][0]

                converter = meta_type_transformers.get(nomad_dtype_str)
                text_value = el.text.strip()  # text representation of incar value
                try:
                    pyvalue = converter(text_value)  # python data type
                except Exception:
                    pyvalue = text_value

                # save (name, pyvalue) into a dict
                self.unknown_incars[name] = pyvalue
                # Unknown_Incars_end
            else:
                if not val_type:
                    val_type = 'float'

                vasp_metainfo_type = vasp_to_metainfo_type_mapping.get(val_type)[0]
                metainfo_type = self._metainfo_type(meta)
                if not vasp_metainfo_type:
                    self.logger.warn("Unknown value type %s encountered in INCAR out: %s %s %r" % (
                        val_type, el.tag, el.attrib, el.text))

                elif metainfo_type != vasp_metainfo_type:
                    if (metainfo_type == 'C' and vasp_metainfo_type == 'b'):
                        pass
                    elif (metainfo_type == 'i' and vasp_metainfo_type == 'f'):
                        pass
                    else:
                        self.logger.warn(
                            "Data type mismatch: %s. Vasp_type: %s, metainfo_type: %s " % (
                                name, vasp_metainfo_type, metainfo_type))
                try:
                    shape = meta.get("shape")
                    converter = meta_type_transformers.get(metainfo_type)
                    if not converter:
                        self.logger.warn(
                            "could not find converter for %s when handling meta info %s" %
                            (metainfo_type, meta))
                    elif shape:
                        vals = re.split(r"\s+", el.text.strip())
                        setattr(run.section_method[-1], meta["name"], [converter(x) for x in vals])
                    else:
                        # If-block to handle incars without value
                        if not el.text:
                            el.text = ''
                        setattr(run.section_method[-1], meta["name"], converter(el.text))

                except Exception:
                    self.logger.warn("Exception trying to handle incarOut %s: %s" % (
                        name, traceback.format_exc()))

                if name == 'ENMAX' or name == 'PREC':
                    if name == 'ENMAX':
                        self.enmax = converter(el.text)
                    if name == 'PREC':
                        if 'acc' in converter(el.text):
                            self.prec = 1.3
                        else:
                            self.prec = 1.0
                if name == 'GGA':
                    f_map = {
                        '91': ['GGA_X_PW91', 'GGA_C_PW91'],
                        'PE': ['GGA_X_PBE', 'GGA_C_PBE'],
                        'RP': ['GGA_X_RPBE', 'GGA_C_PBE'],
                        'PS': ['GGA_C_PBE_SOL', 'GGA_X_PBE_SOL'],
                        'MK': ['GGA_X_OPTB86_VDW'],
                        '--': ['GGA_X_PBE', 'GGA_C_PBE']  # should check potcar
                    }
                    functs = f_map.get(el.text.strip(), None)
                    if not functs:
                        self.logger.warn("Unknown XC functional %s" % el.text.strip())
                    else:
                        for f in functs:
                            section_XC_functionals = run.section_method[-1].m_create(msection_XC_functionals)
                            section_XC_functionals.XC_functional_name = f
                elif name == "ISPIN":
                    self.ispin = int(el.text.strip())

    def _separator_scan(self, element, depth=0):
        for separators in element:
            if separators.tag == "separator":
                for el in separators:
                    if el.tag == "i":
                        self._incar_out_tag(el)
                    elif el.tag == "separator":
                        self._separator_scan(el, depth + 1)
                    else:
                        pass
            elif separators.tag == "i":
                self._incar_out_tag(separators)
            else:
                pass

    def on_end_parameters(self, element, path_str):
        self._separator_scan(element)
        run = self.parser.run
        try:
            self.prec
            try:
                self.enmax
                self.waveCut = run.m_create(msection_basis_set_cell_dependent)
                self.waveCut.basis_set_planewave_cutoff = eV2J(self.enmax * self.prec)

                section_method_basis_set = run.section_method[-1].m_create(msection_method_basis_set)
                section_method_basis_set.mapping_section_method_basis_set_cell_associated = self.waveCut
            except AttributeError:
                import traceback
                traceback.print_exc()
                self.logger.warn(
                    "Missing ENMAX for calculating plane wave basis cut off ")
        except AttributeError:
            self.logger.pwarn(
                "Missing PREC for calculating plane wave basis cut off ")

    def on_end_dos(self, element, path_str):
        "density of states"
        run = self.parser.run
        section_dos = run.section_single_configuration_calculation[-1].m_create(msection_dos)
        for el in element:
            if el.tag == "i":
                if el.attrib.get("name") == "efermi":
                    self.e_fermi = eV2J(float(el.text.strip()))
                    section_dos.energy_reference_fermi = np.array([self.e_fermi] * self.ispin)
                else:
                    self.logger.warn("unexpected tag %s %s in dos" % (el.tag, el.attrib))
            elif el.tag == "total":
                for el1 in el:
                    if el1.tag == "array":
                        for el2 in el1:
                            if el2.tag == "dimension" or el2.tag == "field":
                                pass
                            elif el2.tag == "set":
                                dosL = []
                                for spin_component in el2:
                                    if spin_component.tag == "set":
                                        dosL.append(
                                            get_vector(spin_component, field="r"))
                                    else:
                                        self.logger.warn("unexpected tag %s %s in dos total array set" % (
                                            spin_component.tag, spin_component.attrib))
                                dos_a = np.asarray(dosL)
                                if len(dos_a.shape) != 3:
                                    raise Exception("unexpected shape %s (%s) for total dos (ragged arrays?)" % (
                                        dos_a.shape), dos_a.dtype)
                                dos_e = eV2JV(dos_a[0, :, 0])
                                dos_i = dos_a[:, :, 2]
                                dos_v = dos_a[:, :, 1]

                                # Convert the DOS values to SI. VASP uses the
                                # following units in the output:
                                # states/eV/cell. This means that the volume
                                # dependence has been introduced by multiplying
                                # by the cell volume
                                # the integrated dos value is the number of electrons until that energy level
                                # and thus not directly energy dependent anymore
                                joule_in_ev = convert_unit(1, "eV", "J")
                                dos_v = dos_v / joule_in_ev

                                section_dos.dos_energies = dos_e
                                cell_volume = np.abs(np.linalg.det(self.cell))
                                section_dos.dos_values = dos_v * cell_volume
                                section_dos.dos_integrated_values = dos_i
                            else:
                                self.logger.warn("unexpected tag %s %s in dos total array" % (
                                    el2.tag, el2.attrib))
                    else:
                        self.logger.warn("unexpected tag %s %s in dos total" % (
                            el2.tag, el2.attrib))
            elif el.tag == "partial":
                for el1 in el:
                    if el1.tag == "array":
                        lm = []
                        for el2 in el1:
                            if el2.tag == "dimension":
                                pass
                            elif el2.tag == "field":
                                if el2.text.strip() == "energy":
                                    pass
                                else:
                                    strLm = {
                                        "s": [0, 0],
                                        "p": [1, -1],
                                        "px": [1, 0],
                                        "py": [1, 1],
                                        "pz": [1, 2],
                                        "d": [2, -1],
                                        "dx2": [2, 0],
                                        "dxy": [2, 1],
                                        "dxz": [2, 2],
                                        "dy2": [2, 3],
                                        "dyz": [2, 4],
                                        "dz2": [2, 5]
                                    }
                                    lm.append(
                                        strLm.get(el2.text.strip(), [-1, -1]))
                            elif el2.tag == "set":
                                dosL = []
                                for atom in el2:
                                    if atom.tag == "set":
                                        atomL = []
                                        dosL.append(atomL)
                                        for spin_component in atom:
                                            if spin_component.tag == "set":
                                                atomL.append(
                                                    get_vector(spin_component, field="r"))
                                            else:
                                                self.logger.warn("unexpected tag %s %s in dos partial array set set" % (
                                                    spin_component.tag, spin_component.attrib))
                                    else:
                                        self.logger.warn("unexpected tag %s %s in dos partial array set" % (
                                            spin_component.tag, spin_component.attrib))
                                dosLM = np.asarray(dosL)
                                assert len(
                                    dosLM.shape) == 4, "invalid shape dimension in projected dos (ragged arrays?)"
                                section_dos.dos_values_lm = dosLM[:, :, :, 1:]
                            else:
                                self.logger.warn("unexpected tag %s %s in dos total array" % (
                                    el2.tag, el2.attrib))
                        section_dos.dos_lm = np.asarray(lm)
                        section_dos.dos_m_kind = 'polynomial'
                    else:
                        self.logger.warn("unexpected tag %s %s in dos total" % (
                            el2.tag, el2.attrib))
            else:
                self.logger.warn("unexpected tag %s %s in dos" % (el2.tag, el2.attrib))

    def on_end_projected(self, element, path_str):
        return None


class XmlParser(object):
    @staticmethod
    def extract_callbacks(obj):
        """extracts all callbacks from the object obj

        triggers should start with onStart_ or onEnd__ and then have a valid section name.
        They will be called with this object, the event and current element
        """
        triggers = {}
        for attr in dir(obj):
            if attr.startswith("on_start_"):
                triggers[attr] = getattr(obj, attr)
            elif attr.startswith("on_end_"):
                triggers[attr] = getattr(obj, attr)
        return triggers

    @staticmethod
    def maybe_get(el, path, default=None):
        i = el.findall(path)
        if i:
            return i.pop().text
        else:
            return default

    def __init__(self, parser_info, super_context, metainfo_env=None, callbacks=None, section_map=None):
        self.f_in = None
        self.parser_info = parser_info
        self.super_context = super_context
        self.callbacks = callbacks if callbacks is not None else XmlParser.extract_callbacks(
            super_context)
        self.section_map = section_map if section_map is not None else super_context.section_map
        self.path = []
        self.tagSections = {}
        self.metainfo_env = metainfo_env
        self.run = None
        self.archive = None

    def parse(self, main_file_uri, f_in, archive):
        if self.path:
            raise Exception(
                "Parse of %s called with non empty path, parse already in progress?" % main_file_uri)
        self.main_file_uri = main_file_uri
        self.f_in = f_in
        self.archive = archive
        self.run = archive.m_create(msection_run)
        self.super_context.parser = self
        # there are invalid characters like esc in the files, we do not want to crash on them
        xml_parser = MyXMLParser()
        try:
            for event, el in xml.etree.ElementTree.iterparse(self.f_in, events=["start", "end"], parser=xml_parser):
                if event == 'start':
                    path_str = "/".join([x.tag for x in self.path]) + "/" + el.tag
                    callback = self.callbacks.get("on_start_" + el.tag, None)

                    if callback:
                        callback(el, path_str)
                    self.path.append(el)
                elif event == 'end':
                    last_el = self.path.pop()
                    if last_el != el:
                        raise Exception(
                            "mismatched path at end, got %s expected %s" % (last_el, el))
                    tag = el.tag
                    path_str = "/".join([x.tag for x in self.path]) + "/" + tag
                    callback = self.callbacks.get("on_end_" + tag, None)
                    if callback:
                        if not callback(el, path_str):
                            # if callback does not return True then assume that the current element has been processed
                            # and can be removed
                            el.clear()
                            if self.path:
                                self.path[-1].remove(el)
                    elif len(self.path) == 1:
                        el.clear()
                        self.path[-1].remove(el)
                else:
                    raise Exception("Unexpected event %s" % event)
        except ParseError as e:
            self.super_context.logger.error("Could not complete parsing: %s" % e, exc_info=e)
        except Exception as e:
            import traceback
            traceback.print_exc()
        else:
            pass


g = XmlParser.maybe_get

parser_info = {
    "name": "parser_vasprun",
    "version": "1.0"
}



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ase.data import atomic_names, atomic_numbers, chemical_symbols,\
        atomic_masses

def name(self):
    return self._name


def atomicNumber(self):
    return self._atomic_number


def symbol(self):
    return self._symbol


def atomicMass(self):
    return self._atomic_mass_amu

things = {}
for n, s, m in zip(atomic_names[1:], chemical_symbols[1:],
                      atomic_masses[1:]):
    stuff = {'_atomic_number': atomic_numbers[s],
             '_name': n,
             '_symbol': s,
             '_atomic_mass_amu': m,
             'atomicNumber': atomicNumber,
             'name': name,
             'atomicMass': atomicMass,
             'symbol': symbol}
    element = type(n, (object,), stuff)
    exec(n + ' = element()')
    #exec(s + ' = ' + n)
    exec('things[n] = ' + n)

# clean up
del stuff, name, atomicNumber, symbol, atomicMass, n, s, m, element
del atomic_names, atomic_numbers, chemical_symbols, atomic_masses



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json
import datetime

from nomad import infrastructure, datamodel, utils

from .admin import admin


@admin.group(help='''Add, import, export users.''')
def users():
    pass


@users.command(help='Import users to keycloak from a JSON file.', name='import')
@click.argument('PATH_TO_USERS_FILE', type=str, nargs=1)
def import_command(path_to_users_file):
    with open(path_to_users_file, 'rt') as f:
        users = json.load(f)

    logger = utils.get_logger(__name__)

    for user_dict in users:
        try:
            password = user_dict.pop('password')
            user_dict['created'] = datetime.datetime.fromtimestamp(user_dict['created'] / 1000)
            user = datamodel.User(**user_dict)
            infrastructure.keycloak.add_user(user, bcrypt_password=password, invite=False)
            print('Imported %s' % user.name)
        except Exception as e:
            logger.error('could not import user', exc_info=e)



# Copyright 2016-2018 The NOMAD Developers Group
# Copyright 2017-2018 Lorenzo Pardini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xml.sax
import logging
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.unit_conversion import unit_conversion


class InputHandler(xml.sax.handler.ContentHandler):
    def __init__(self, backend):
        self.backend = backend

    def startElement(self, name, attrs):
        if name == "libxc":        #libXC
            correlation = attrs.getValue("correlation")[3:]
            exchange = attrs.getValue("exchange")[3:]
            xcName = [correlation, exchange]
            for xc in xcName:
                gi = self.backend.openSection("section_XC_functionals")
                self.backend.addValue("XC_functional_name", xc)
                self.backend.closeSection("section_XC_functionals", gi)

    def endElement(self, name):
        pass

def parseInput(inF, backend):
    handler = InputHandler(backend)
    logging.info("will parse")
    xml.sax.parse(inF, handler)
    logging.info("did parse")
